#include <smallgmp.h>
#include <stuff_64.h>

#ifndef ABS
#define ABS(x) (x >= 0 ? x : -x)
#endif

#ifndef BITS_PER_BYTE
#define BITS_PER_BYTE   8
#endif
#define MIN_ALLOC         2
#define BYTES_MIN_ALLOC   8
#ifdef SIZEOF_UNSIGNED_INT
#define BYTES_PER_MP_LIMB SIZEOF_UNSIGNED_INT
#else
#define BYTES_PER_MP_LIMB 4
#endif
const int mp_bits_per_limb = BYTES_PER_MP_LIMB*BITS_PER_BYTE;

#define NUM_BASIS 100000000

static void * _mp_default_allocate (size_t);
static void * _mp_default_reallocate (void *, size_t, size_t);
static void _mp_default_free (void *, size_t);
static int  mpn_cmp (mp_srcptr, mp_srcptr, int);
static void mpn_a_s (MP_INT *, const MP_INT *, const MP_INT *, const int);
static void nlAddMult (mp_ptr, mp_srcptr, const int, const unsigned int);
static void nlIniMult (mp_ptr, mp_srcptr, const int, const unsigned int);
static mp_limb_t nlDivModSmall (mp_ptr, mp_srcptr, int, const unsigned int);
static int nlDivModLong (mp_ptr, const int, mp_ptr, int, mp_ptr, const int);
static mp_limb_t nlRemSmall (mp_srcptr, int, const unsigned int);
static int  nlRemLong(mp_ptr, mp_ptr, int, const int);
static void nlGcdLong (mp_ptr, int *, mp_ptr, int, mp_ptr *);
static void mpn_gcd (MP_INT *, const MP_INT *, const MP_INT *);

void mpz_gcdext (MP_INT *g, MP_INT *s, MP_INT *t, const MP_INT *a, const MP_INT *b)
{
  MP_INT s0, s1, q, r, x, d0, d1;

  mpz_init (&s0);
  mpz_init (&s1);
  mpz_init (&q);
  mpz_init (&r);
  mpz_init (&x);
  mpz_init_set (&d0, a);
  mpz_init_set (&d1, b);
  mpz_set_ui (&s0, 1L);
  mpz_set_ui (&s1, 0L);

  while (d1._mp_size != 0)
    {
      mpz_tdiv_qr (&q, &r, &d0, &d1);
      mpz_set (&d0, &d1);
      mpz_set (&d1, &r);

      mpz_mul (&x, &s1, &q);
      mpz_sub (&x, &s0, &x);
      mpz_set (&s0, &s1);
      mpz_set (&s1, &x);
    }

  if (t != NULL)
    {
      mpz_mul (&x, &s0, a);
      mpz_sub (&x, &d0, &x);
      if (b->_mp_size == 0)
	    t->_mp_size = 0;
      else
	    mpz_tdiv_q (t, &x, b);
    }
  mpz_set (s, &s0);
  mpz_set (g, &d0);
  if (g->_mp_size < 0)
    {
      g->_mp_size = -g->_mp_size;
      s->_mp_size = -s->_mp_size;
      if (t != NULL)
	    t->_mp_size = -t->_mp_size;
    }

  mpz_clear (&s0);
  mpz_clear (&s1);
  mpz_clear (&q);
  mpz_clear (&r);
  mpz_clear (&x);
  mpz_clear (&d0);
  mpz_clear (&d1);
}

/********************************************/
/**************** set memory ****************/


#ifndef SMALLGMP_DEBUG
void *(*_mp_allocate_func) (size_t)=_mp_default_allocate;
void *(*_mp_reallocate_func) (void *, size_t, size_t)=_mp_default_reallocate;
void (*_mp_free_func) (void *, size_t)=_mp_default_free;

void mp_set_memory_functions (void *(*alloc_func) (size_t),
void *(*realloc_func) (void *, size_t, size_t),
void (*free_func) (void *, size_t)
)
{
  _mp_allocate_func = alloc_func;
  _mp_reallocate_func = realloc_func;
  _mp_free_func = free_func;
}


#else
void *(*_mp_allocate_func) (size_t, char*, int);
void *(*_mp_reallocate_func) (void *, size_t, size_t, char*, int);
void (*_mp_free_func) (void *, size_t, char*, int);

void mp_set_memory_functions (void *(*alloc_func) (size_t, char*, int),
void *(*realloc_func) (void *, size_t, size_t, char*, int),
void (*free_func) (void *, size_t, char*, int)
)
{
  _mp_allocate_func = alloc_func;
  _mp_reallocate_func = realloc_func;
  _mp_free_func = free_func;
}


#endif
/********************************************/
/******************* misc *******************/

/* x := 0 (x->_mp_d is free) */
void mpz_init (MP_INT *x)
{
  x->_mp_alloc = MIN_ALLOC;
  x->_mp_size = 0;
  x->_mp_d = (mp_ptr) mmpAlloc (BYTES_MIN_ALLOC);
}


void _mpz_realloc (MP_INT *x, int new_size) /* !! void * !! */
{
  if (new_size > x->_mp_alloc) {
    mp_ptr bl = x->_mp_d;
    int i = ABS(x->_mp_size) - 1;
    mp_ptr al = (mp_ptr) mmpAlloc (BYTES_PER_MP_LIMB * new_size);
    for (; i >= 0; i--)
      al[i] = bl[i];
    mmpFree (bl, BYTES_PER_MP_LIMB * x->_mp_alloc);
    x->_mp_alloc = new_size;
    x->_mp_d = al;
  }
}


/* dst := src (dst->_mp_d is free) */
void mpz_init_set (MP_INT *dst, const MP_INT *src)
{
  const int size = src->_mp_size;
  int i = ABS(size);
  mp_srcptr bl = src->_mp_d;
  mp_ptr al;

  if (i <= MIN_ALLOC) {
    al = (mp_ptr) mmpAlloc (BYTES_MIN_ALLOC);
    dst->_mp_alloc = MIN_ALLOC;
  } else {
    al = (mp_ptr) mmpAlloc (BYTES_PER_MP_LIMB * i);
    dst->_mp_alloc = i;
  }
  i--;
  for (; i >= 0; i--)
    al[i] = bl[i];
  dst->_mp_size = size;
  dst->_mp_d = al;
}


/* dst := src (dst->_mp_d has memory) */
void mpz_set (MP_INT *dst, const MP_INT *src)
{
  const int size = src->_mp_size;
  int i = ABS(size);
  mp_srcptr bl = src->_mp_d;
  mp_ptr al = dst->_mp_d;

  if (al != bl) {
    if (dst->_mp_alloc < i) {
      mmpFree (al, BYTES_PER_MP_LIMB * dst->_mp_alloc);
      al = dst->_mp_d = (mp_ptr) mmpAlloc (BYTES_PER_MP_LIMB * i);
      dst->_mp_alloc = i;
    }

    dst->_mp_size = size;
    i--;
    for (; i >= 0; i--)
      al[i] = bl[i];
  }
}

void mpz_init_set_ui (MP_INT *dst, unsigned long int abs_val)
{
  mp_ptr al = (mp_ptr) mmpAlloc (BYTES_MIN_ALLOC);

  dst->_mp_alloc = MIN_ALLOC;
  dst->_mp_d = al;
  if (abs_val == 0) {
    dst->_mp_size = 0;
    return;
  }
  al[0] = abs_val;
  dst->_mp_size = 1;
}

/* dst := val (dst->_mp_d is free) */
void mpz_init_set_si (MP_INT *dst, signed long int val)
{
  mp_ptr al = (mp_ptr) mmpAlloc (BYTES_MIN_ALLOC);

  dst->_mp_alloc = MIN_ALLOC;
  dst->_mp_d = al;
  if (val == 0) {
    dst->_mp_size = 0;
    return;
  }
  if (val > 0) {
    al[0] = val;
    dst->_mp_size = 1;
  } else {
    al[0] = -val;
    dst->_mp_size = -1;
  }
}

/* dst := val (dst->_mp_d has memory) */
void mpz_set_ui (MP_INT *dst, unsigned long int uval)
{
  if (uval == 0) {
    dst->_mp_size = 0;
    return;
  }

  dst->_mp_d[0] = uval;
  dst->_mp_size = 1;
}

/* dst := val (dst->_mp_d has memory) */
void mpz_set_si (MP_INT *dst, signed long int val)
{
  if (val == 0) {
    dst->_mp_size = 0;
    return;
  }

  if (val > 0) {
    dst->_mp_d[0] = val;
    dst->_mp_size = 1;
  } else {
    dst->_mp_d[0] = -val;
    dst->_mp_size = -1;
  }
}


signed long int mpz_get_si (const MP_INT *src)
{
  const int size = src->_mp_size;
  mp_limb_t i = src->_mp_d[0];

    if (size > 0)
      return i&0X8FFFFFFF;
    else if (size < 0)
      return ~((i-1)&0X8FFFFFFF);
    else
      return 0;
}

unsigned long int mpz_get_ui (const MP_INT *a)
{
  if(a->_mp_size)
    return a->_mp_d[0];
  else
    return 0;
}

void mpz_abs (MP_INT *w, const MP_INT *u)
{
  mpz_set(w,u);
  if(w->_mp_size<0)
    w->_mp_size = -w->_mp_size;
}

/* make x->_mp_d free */
void mpz_clear (MP_INT *x)
{
  mmpFree (x->_mp_d, BYTES_PER_MP_LIMB * x->_mp_alloc);
}


void mpz_set_str (MP_INT *dst, const char *str, int base) /* !! int !! */
{
  if (*str == '\0') {
    dst->_mp_size = 0;
    return;
  }
  dst->_mp_d[0] = (mp_limb_t) * str - (mp_limb_t)'0';
  dst->_mp_size = 1;
  str++;
  while (*str != '\0') {
    mpz_mul_ui(dst, dst, base);
    mpz_add_ui(dst, dst, (mp_limb_t)*str - (mp_limb_t)'0');
    str++;
  }
}


void mpz_init_set_str (MP_INT *dst, const char *str, int base) /* !! int !! */
{
  mpz_init(dst);
  if (*str == '\0') {
    dst->_mp_size = 0;
    return;
  }
  dst->_mp_d[0] = (mp_limb_t) * str - (mp_limb_t)'0';
  dst->_mp_size = 1;
  str++;
  while (*str != '\0') {
    mpz_mul_ui(dst, dst, base);
    mpz_add_ui(dst, dst, (mp_limb_t)*str - (mp_limb_t)'0');
    str++;
  }
}


size_t mpz_sizeinbase (const MP_INT *x, int base)
{
  return (size_t) ((10 * ABS(x->_mp_size)) + 8);
}


char  *mpz_get_str (char *str, int base, const MP_INT *src)
{
  MP_INT b;
  mp_ptr a;
  int al;
  char  *str1;
  char  save;
  mp_limb_t w;

  mpz_init_set(&b, src);
  a = b._mp_d;
  str1 = str + (mpz_sizeinbase(&b, base) + 1);
  *str1 = '\0';
  al = ABS(src->_mp_size);
  do {
    w = nlDivModSmall(a, a, al, NUM_BASIS);
    save = *str1;
    str1 -= 8;
    sprintf(str1, "%08u", w);
    str1[8] = save;
    if (a[al-1] == 0)
      al--;
  } while (al > 0);
  mpz_clear(&b);
  while (*str1 == '0') 
    str1++;
  if (src->_mp_size < 0) {
    str1--;
    *str1 = '-';
  }
  return str1;
}


/********************************************/
/****************** compare *****************/

int  mpz_cmp (const MP_INT *u, const MP_INT *v)
{
  const int usize = u->_mp_size;
  const int vsize = v->_mp_size;
  mp_ptr up, vp;
  int i;
  mp_limb_t u_val, v_val;

  if (usize != vsize)
    return usize - vsize;
  if (usize == 0)
    return 0;
    
  up = u->_mp_d;
  vp = v->_mp_d;

  for (i = ABS(usize) - 1; i >= 0; i--) {
    u_val = up[i];
    v_val = vp[i];
    if (u_val != v_val)
      goto diff;
  }
  return 0;

diff:
  if (usize > 0)
    return (u_val > v_val) ? 1 : -1;
  else
    return (u_val > v_val) ? -1 : 1;
}


int  mpz_cmp_ui (const MP_INT *u, unsigned long int v_digit)
{
  const int usize = u->_mp_size;
  mp_limb_t u_val;

  if (v_digit == 0)
    return usize;

  if (usize != 1)
    return usize - 1;

  u_val = u->_mp_d[0];
  if (u_val != v_digit)
    return (u_val > v_digit) ? 1 : -1;
  return 0;
}


int  mpz_cmp_si (const MP_INT *u, signed long int v_digit)
{
  const int usize = u->_mp_size;
  mp_limb_t u_val;

  if (v_digit == 0)
    return usize;

  if (v_digit > 0)
  {
    if (usize != 1)
      return usize - 1;
  }
  else
  {
    if (usize != -1)
      return usize + 1;
    v_digit = -v_digit;
  }

  u_val = u->_mp_d[0];
  if (u_val != v_digit)
  {
    if (usize > 0)
      return (u_val > v_digit) ? 1 : -1;
    else
      return (u_val > v_digit) ? -1 : 1;
  }
  else
    return 0;
}


/********************************************/
/**************** operations ****************/

void mpz_neg (MP_INT *dst, const MP_INT *src)
{
  const int src_size = src->_mp_size;

  if (src != dst) {
    mp_ptr al = dst->_mp_d, bl = src->_mp_d;
    int i = ABS (src_size);
    if (dst->_mp_alloc < i) {
      mmpFree (al, BYTES_PER_MP_LIMB * dst->_mp_alloc);
      al = dst->_mp_d = (mp_ptr) mmpAlloc (BYTES_PER_MP_LIMB * i);
      dst->_mp_alloc = i;
    }
    i--;
    for (; i >= 0; i--)
      al[i] = bl[i];
  }
  dst->_mp_size = -(src_size);
}


void mpz_add (MP_INT *sum, const MP_INT *u, const MP_INT *v)
{
  mpn_a_s(sum, u, v, 0);
}


void mpz_sub (MP_INT *dif, const MP_INT *u, const MP_INT *v)
{
  mpn_a_s(dif, u, v, 1);
}


void mpz_add_ui (MP_INT *sum, const MP_INT *add1, unsigned long int add2)
{
  MP_INT tmp;

  if (add2 == 0) {
    mpz_set(sum, add1);
    return;
  }
  tmp._mp_alloc = MIN_ALLOC;
  tmp._mp_d = (mp_ptr) mmpAlloc (BYTES_MIN_ALLOC);
  tmp._mp_d[0] = add2;
  tmp._mp_size = 1;
  mpn_a_s(sum, add1, &tmp, 0);
  mmpFree(tmp._mp_d, BYTES_MIN_ALLOC);
}


void mpz_sub_ui (MP_INT *dif, const MP_INT *min, unsigned long int sub)
{
  MP_INT tmp;

  if (sub == 0) {
    mpz_set(dif, min);
    return;
  }
  tmp._mp_alloc = MIN_ALLOC;
  tmp._mp_d = (mp_ptr) mmpAlloc (BYTES_MIN_ALLOC);
  tmp._mp_d[0] = sub;
  tmp._mp_size = 1;
  mpn_a_s(dif, min, &tmp, 1);
  mmpFree(tmp._mp_d, BYTES_MIN_ALLOC);
}


void mpz_mul (MP_INT *w, const MP_INT *u, const MP_INT *v)
{
  int delalloc = 0;
  int usize = u->_mp_size;
  int vsize = v->_mp_size;
  int wsize;
  int sign_product, t;
  mp_ptr up, vp, wp, del;

  sign_product = usize ^ vsize;
  usize = ABS (usize);
  vsize = ABS (vsize);

  if (usize < vsize) {
    up = v->_mp_d; 
    vp = u->_mp_d;
    t = usize; 
    usize = vsize; 
    vsize = t;
  } else {
    up = u->_mp_d; 
    vp = v->_mp_d;
  }
  if (vsize == 0) {
    w->_mp_size = 0;
    return;
  }

  /* Ensure W has space enough to store the result.  */
  wp = w->_mp_d;
  wsize = usize + vsize;
  if ((wp == up) || (wp == vp)) {
    del = wp;
    delalloc = w->_mp_alloc;
    if (wsize <= MIN_ALLOC) {
      w->_mp_alloc = MIN_ALLOC;
      wp = w->_mp_d = 
          (mp_ptr) mmpAlloc (BYTES_MIN_ALLOC);
    } else {
      w->_mp_alloc = wsize;
      wp = w->_mp_d = 
          (mp_ptr) mmpAlloc (BYTES_PER_MP_LIMB * wsize);
    }
  } else if (w->_mp_alloc < wsize) {
    mmpFree (wp, BYTES_PER_MP_LIMB * w->_mp_alloc);
    wp = w->_mp_d = 
        (mp_ptr) mmpAlloc (BYTES_PER_MP_LIMB * wsize);
    w->_mp_alloc = wsize;
  }

  nlIniMult(wp, up, usize, vp[0]);
  for (t = 1; t < vsize; t++)
    nlAddMult(wp + t, up, usize, vp[t]);
  if (wp[wsize - 1] == 0)
    wsize--;
  w->_mp_size = sign_product < 0 ? -wsize : wsize;
  if (delalloc != 0)
    mmpFree (del, delalloc * BYTES_PER_MP_LIMB);
}


void mpz_mul_ui (MP_INT *w, const MP_INT *u,
unsigned long int small_mult)
{
  int delalloc = 0;
  const int usize = ABS(u->_mp_size);
  mp_srcptr up = u->_mp_d;
  int wsize;
  mp_ptr wp, del;

  if ((usize == 0) || (small_mult == 0)) {
    w->_mp_size = 0;
    return;
  }
  wp = w->_mp_d;
  wsize = usize + 1;

  /* Ensure W has space enough to store the result.  */
  if (wp == up) {
    del = wp;
    delalloc = w->_mp_alloc;
    if (wsize <= MIN_ALLOC) {
      w->_mp_alloc = MIN_ALLOC;
      wp = w->_mp_d = 
          (mp_ptr) mmpAlloc (BYTES_MIN_ALLOC);
    } else {
      w->_mp_alloc = wsize;
      wp = w->_mp_d = 
          (mp_ptr) mmpAlloc (BYTES_PER_MP_LIMB * wsize);
    }
  } else if (w->_mp_alloc < wsize) {
    mmpFree (wp, BYTES_PER_MP_LIMB * w->_mp_alloc);
    wp = w->_mp_d = 
        (mp_ptr) mmpAlloc (BYTES_PER_MP_LIMB * wsize);
    w->_mp_alloc = wsize;
  }

  nlIniMult(wp, up, usize, small_mult);
  if (wp[wsize - 1] == 0)
    wsize--;
  w->_mp_size = u->_mp_size < 0 ? -wsize : wsize;
  if (delalloc != 0)
    mmpFree (del, delalloc * BYTES_PER_MP_LIMB);
}


void mpz_tdiv_qr (MP_INT *quot, MP_INT *rem, const MP_INT *num, const MP_INT *den)
{
  const int sign_remainder = num->_mp_size;
  int dsize = den->_mp_size;
  const int sign_quotient = sign_remainder ^ dsize;
  mp_ptr del1 = NULL, del2 = NULL, dp = den->_mp_d;
  mp_ptr rp, qp;
  int del1alloc, del2alloc, rsize, qsize, t;

  if ((sign_remainder == 0) || (dsize == 0)) {
    quot->_mp_size = 0;
    rem->_mp_size = 0;
    return;
  }
  dsize = ABS (dsize);

  if (rem->_mp_d == dp) {
    del1 = dp;
    del1alloc = den->_mp_alloc;
    mpz_init_set(rem, num);
  }
  else
    mpz_set(rem, num);
  rp = rem->_mp_d;
  rsize = ABS (sign_remainder);

  t = qsize = rsize - dsize;
  if (t == 0)
    t = mpn_cmp(rp, dp, dsize);

  if (t <= 0) {
    if (t < 0)
      quot->_mp_size = 0;
    else {
      if (sign_quotient < 0)
        quot->_mp_size = -1;
      else
        quot->_mp_size = 1;
      quot->_mp_d[0] = 1;
      rem->_mp_size = 0;
    }
    if (del1 != NULL)
      mmpFree (del1, del1alloc * BYTES_PER_MP_LIMB);
    return;
  }

  if (qsize > 0) {
    t = mpn_cmp(rp + qsize, dp, dsize);
  }
  if (t >= 0)
    qsize++;

  qp = quot->_mp_d;
  if ((qp == rp) || (qp == dp)) {
    del2 = qp;
    del2alloc = quot->_mp_alloc;
    if (qsize <= MIN_ALLOC) {
      quot->_mp_alloc = MIN_ALLOC;
      qp = quot->_mp_d = 
          (mp_ptr) mmpAlloc (BYTES_MIN_ALLOC);
    } else {
      quot->_mp_alloc = qsize;
      qp = quot->_mp_d = 
          (mp_ptr) mmpAlloc (BYTES_PER_MP_LIMB * qsize);
    }
  } else if (quot->_mp_alloc < qsize) {
    mmpFree (qp, BYTES_PER_MP_LIMB * quot->_mp_alloc);
    qp = quot->_mp_d = 
        (mp_ptr) mmpAlloc (BYTES_PER_MP_LIMB * qsize);
    quot->_mp_alloc = qsize;
  }

  if (dsize == 1) {
    rp[0] = nlDivModSmall(qp, rp, rsize, dp[0]);
    if (rp[0] != 0)
      t = 1;
    else
      t = 0;
  } else
    t = nlDivModLong(qp, qsize, rp, rsize, dp, dsize);
  if (sign_remainder < 0)
    t = -t;
  rem->_mp_size = t;
  if (sign_quotient < 0)
    qsize = -qsize;
  quot->_mp_size = qsize;

  if (del1 != NULL)
    mmpFree (del1, del1alloc * BYTES_PER_MP_LIMB);
  if (del2 != NULL)
    mmpFree (del2, del2alloc * BYTES_PER_MP_LIMB);
}

void mpz_fdiv_qr (MP_INT *q, MP_INT *r, const MP_INT *a, const MP_INT *b)
{
  mpz_tdiv_qr(q, r, a, b);
  if ((a->_mp_size ^ b->_mp_size) < 0 && r->_mp_size != 0)
  {
    mpz_sub_ui (q, q, 1L);
    mpz_add (r, r, b);
  }
}

void mpz_cdiv_qr (MP_INT *q, MP_INT *r, const MP_INT *a, const MP_INT *b)
{
  mpz_tdiv_qr(q, r, a, b);
  if ((a->_mp_size ^ b->_mp_size) >= 0 && r->_mp_size != 0)
  {
    mpz_add_ui (q, q, 1L);
    mpz_sub (r, r, b);
  }
}

void mpz_tdiv_q (MP_INT *quot, const MP_INT *num, const MP_INT *den)
{
  MP_INT b, *rem = &b;

  mpz_init(rem);
  mpz_tdiv_qr(quot, rem, num, den);
  mpz_clear(rem);
}

void mpz_fdiv_q (MP_INT *quot, const MP_INT *num, const MP_INT *den)
{
  MP_INT b, *rem = &b;

  mpz_init(rem);
  mpz_tdiv_qr(quot, rem, num, den);
  if ((num->_mp_size ^ den->_mp_size) < 0 && rem->_mp_size != 0)
  {
    mpz_sub_ui (quot, quot, 1L);
  }
  mpz_clear(rem);
}

void mpz_cdiv_q (MP_INT *quot, const MP_INT *num, const MP_INT *den)
{
  MP_INT b, *rem = &b;

  mpz_init(rem);
  mpz_tdiv_qr(quot, rem, num, den);
  if ((num->_mp_size ^ den->_mp_size) >= 0 && rem->_mp_size != 0)
  {
    mpz_add_ui (quot, quot, 1L);
  }
  mpz_clear(rem);
}

void mpz_tdiv_r (MP_INT *rem, const MP_INT *num, const MP_INT *den)
{
  const int sign_remainder = num->_mp_size;
  int dsize = den->_mp_size;
  mp_ptr del = NULL, dp = den->_mp_d;
  mp_ptr rp;
  int delalloc, rsize, t;

  if ((sign_remainder == 0) || (dsize == 0)) {
    rem->_mp_size = 0;
    return;
  }
  dsize = ABS (dsize);

  if (rem->_mp_d == dp) {
    del = dp;
    delalloc = den->_mp_alloc;
    mpz_init_set(rem, num);
  }
  else
    mpz_set(rem, num);
  rp = rem->_mp_d;
  rsize = ABS (sign_remainder);

  t = rsize - dsize;
  if (t == 0) {
    t = mpn_cmp(rp, dp, dsize);
    if (t == 0)
      rem->_mp_size = 0;
  }

  if (t > 0) {
    if (dsize == 1) {
      rp[0] = (mp_limb_t)nlRemSmall(rp, rsize, dp[0]);
      t = 1;
      if (rp[0] == 0)
        t = 0;
    } else
      t = nlRemLong(rp, dp, rsize, dsize);
    if (sign_remainder < 0)
      t = -t;
    rem->_mp_size = t;
  }
  if (del != NULL)
    mmpFree (del, delalloc * BYTES_PER_MP_LIMB);
}

void mpz_mod (MP_INT *rem, const MP_INT *num, const MP_INT *den)
{
  mpz_tdiv_r(rem, num, den);
  if (rem->_mp_size != 0)
  {
    if (num->_mp_size < 0)
      if (den->_mp_size < 0)
        mpz_sub (rem, rem, den);
      else
        mpz_add (rem, rem, den);
  }
}

unsigned long int  mpz_fdiv_ui (
const MP_INT *dividend, unsigned long int divisor)
{
  int ds = dividend->_mp_size, dt = ABS(ds);
  mp_ptr a = dividend->_mp_d;
  unsigned long int  res;

  if ((dt == 0) || (divisor == 0))
    return 0;
  if ((dt < 2) || (divisor < 3)) {
    res = a[0]%divisor;
    if ((ds < 0) && res)
      res = divisor - res;
    return res;
  }
  res = nlRemSmall(a, dt, divisor);
  if ((ds < 0) && res)
    res = divisor - res;
  return res;
}

unsigned long int  mpz_cdiv_ui (
const MP_INT *dividend, unsigned long int divisor)
{
  int ds = dividend->_mp_size, dt = ABS(ds);
  mp_ptr a = dividend->_mp_d;
  unsigned long int  res;

  if ((dt == 0) || (divisor == 0))
    return 0;
  if ((dt < 2) || (divisor < 3)) {
    res = a[0]%divisor;
    if ((ds < 0) && res)
      res = divisor - res;
    return res;
  }
  res = nlRemSmall(a, dt, divisor);
  if ((ds >= 0) && res)
    res = divisor - res;
  return res;
}

void  mpz_fdiv_r_ui (MP_INT *rem,
const MP_INT *dividend, unsigned long int divisor)
{
  unsigned long int  res = mpz_fdiv_ui(dividend, divisor);

  if (res) {
    if (dividend->_mp_size < 0)
      res = divisor - res;
    rem->_mp_d[0] = res;
    rem->_mp_size = 1;
  }
  else
    rem->_mp_size = 0;
}

unsigned long int mpz_fdiv_q_ui (MP_INT *q, const MP_INT *a, unsigned long int b)
{
  MP_INT bi, ri;
  unsigned long int res;

  if (b && a->_mp_size)
  {
    mpz_init(&ri);
    mpz_init_set_ui(&bi,b);
    mpz_tdiv_qr(q, &ri, a, &bi);
    if (ri._mp_size != 0)
    {
      res = (long)ri._mp_d[0];
      if (a->_mp_size < 0)
      {
        mpz_sub_ui (q, q, 1L);
        res = b - res;
      }
    }
    else
      res = 0;
    mpz_clear(&bi);
    mpz_clear(&ri);
    return res;
  }
  else
  {
    q->_mp_size = 0;
    return 0;
  }
}

unsigned long int mpz_fdiv_qr_ui (MP_INT *q, MP_INT *r, const MP_INT *a, unsigned long int b)
{
  MP_INT bi;
  unsigned long int res;

  if (b && a->_mp_size)
  {
    mpz_init_set_ui(&bi,b);
    mpz_tdiv_qr(q, r, a, &bi);
    if (r->_mp_size != 0)
    {
      res = (long)r->_mp_d[0];
      if (a->_mp_size < 0)
      {
        mpz_sub_ui (q, q, 1L);
        res = b - res;
        r->_mp_d[0] = res;
      }
    }
    else
      res = 0;
    mpz_clear(&bi);
    return res;
  }
  else
  {
    q->_mp_size = 0;
    r->_mp_size = 0;
    return 0;
  }
}

void mpz_gcd (MP_INT *w, const MP_INT *u, const MP_INT *v)
{
  int usize = ABS (u->_mp_size), vsize = ABS (v->_mp_size);
  int i = usize - vsize;

  if (i == 0) {
    i = mpn_cmp(u->_mp_d, v->_mp_d, usize);
    if (i == 0) {
      if ((w->_mp_d != u->_mp_d) && (w->_mp_d != v->_mp_d))
        mpz_set(w, u);
      w->_mp_size = usize;
      return;
    }
  }
  if (i < 0)
    mpn_gcd(w, v, u);
  else
    mpn_gcd(w, u, v);
}

unsigned long int
mpz_gcd_ui (MP_INT *w, const MP_INT *u, unsigned long int v)
{
  MP_INT vi, *t;
  unsigned long int r;

  mpz_init_set_ui(&vi,v);
  if (w) t = w;
  else t = &vi;
  mpz_gcd(t,u,&vi);
  r = t->_mp_d[0];
  mpz_clear(&vi);
  return r;
}

void mpz_pow_ui (MP_INT *r, const MP_INT *b, unsigned long int e)
{
  MP_INT y;

  if (e == 0) {
    r->_mp_d[0] = 1;
    r->_mp_size = 1;
    return;
  }
  mpz_init_set_si(&y, (long)1);
  mpz_set(r, b);
  do {
    if (e % 2 != 0)
      mpz_mul(&y, &y, r);
    if (e > 1)
      mpz_mul(r, r, r);
    e /= 2;
  } while (e);
  mpz_set(r, &y);
}

void mpz_sqrt (MP_INT *sqr, const MP_INT *a)
{
  int i = a->_mp_size;

  if (i < 0)
  {
    printf("***** sqrt ******\n");
    sqr->_mp_size = 0;
    return;
  }
  mpz_set(sqr, a);
  if (i != 0)
  {
    i--;
    sqr->_mp_d[i] /= 2;
    if (sqr->_mp_d[i] == 0)
      sqr->_mp_size = i;
  }
}


static mp_limb_t nlSubSimple (mp_ptr, mp_srcptr, const int);
static mp_limb_t nlSubMult (mp_ptr, mp_srcptr, const unsigned int, const int);
static void nlLShift(mp_ptr, const unsigned int, const unsigned int);
static void nlRShift(mp_ptr, const unsigned int, const unsigned int);
static int nlCompHigh (mp_srcptr, mp_srcptr, const int);
static int  nlShrink0(mp_srcptr, int);
static int  mpn_sub (mp_ptr, mp_srcptr, int, mp_srcptr, int);
static int  mpn_add (mp_ptr, mp_srcptr, int, mp_srcptr, int);

static void * _mp_default_allocate (size_t size)
{
  return malloc (size);
}

static void * _mp_default_reallocate (void *oldptr, size_t old_size, size_t new_size)
{
  return realloc (oldptr, new_size);
}

static void _mp_default_free (void *blk_ptr, size_t blk_size)
{
  free (blk_ptr);
}

static int  mpn_cmp (mp_srcptr op1_ptr, mp_srcptr op2_ptr, int size)
{
  int i;
  mp_limb_t op1_word, op2_word;

  for (i = size - 1; i >= 0; i--) {
    op1_word = op1_ptr[i];
    op2_word = op2_ptr[i];
    if (op1_word != op2_word)
      goto diff;
  }
  return 0;
diff:
  return (op1_word > op2_word) ? 1 : -1;
}

static void mpn_a_s (MP_INT *w, const MP_INT *u, const MP_INT *v, const int c)
{
  int abs_usize = ABS (u->_mp_size), abs_vsize = ABS (v->_mp_size);
  int delalloc = 0;
  int usize, vsize, wsize, t;
  mp_ptr up, vp, wp, del;

  if (abs_usize < abs_vsize) {
    t = abs_usize; 
    abs_usize = abs_vsize; 
    abs_vsize = t;
    if (c != 0)
      usize = -v->_mp_size;
    else
      usize = v->_mp_size;
    vsize = u->_mp_size;
    up = v->_mp_d; 
    vp = u->_mp_d;
  } else {
    if (abs_usize == 0) {
      w->_mp_size = 0;
      return;
    }
    if (c != 0)
      vsize = -v->_mp_size;
    else
      vsize = v->_mp_size;
    usize = u->_mp_size;
    up = u->_mp_d; 
    vp = v->_mp_d;
  }

  /* If not space for w (and possible carry), increase space.
     substitute _mpz_realloc */
  wp = w->_mp_d;
  wsize = abs_usize;
  if (usize > 0)
  {
    if (vsize > 0)
      wsize++;
  }
  else
  {
    if (vsize < 0)
      wsize++;
  }
  if (w->_mp_alloc < wsize) {
    delalloc = w->_mp_alloc;
    del = wp;
    wp = w->_mp_d = (mp_ptr) mmpAlloc (BYTES_PER_MP_LIMB * wsize);
    w->_mp_alloc = wsize;
  }

  if (usize > 0) {
    if (vsize >= 0) {
      wsize = mpn_add (wp, up, abs_usize, vp, abs_vsize);
    } else {
      /* The signs are different.  Need exact comparision to determine
         which operand to subtract from which.  */
      if (abs_usize != abs_vsize) {
        wsize = mpn_sub (wp, up, abs_usize, vp, abs_vsize);
      } else {
        t = mpn_cmp (up, vp, abs_usize);
        if (t > 0) {
          wsize = mpn_sub (wp, up, abs_usize, vp, abs_usize);
        } else if (t < 0) {
          wsize = -mpn_sub (wp, vp, abs_usize, up, abs_usize);
        } else
          wsize = 0;
      }
    }
  } else {
    if (vsize > 0) {
      /* The signs are different.  Need exact comparision to determine
         which operand to subtract from which.  */
      if (abs_usize != abs_vsize) {
        wsize = -mpn_sub (wp, up, abs_usize, vp, abs_vsize);
      } else {
        t = mpn_cmp (up, vp, abs_usize);
        if (t > 0) {
          wsize = -mpn_sub (wp, up, abs_usize, vp, abs_usize);
        } else if (t < 0) {
          wsize = mpn_sub (wp, vp, abs_usize, up, abs_usize);
        } else
          wsize = 0;
      }
    } else {
      wsize = -mpn_add (wp, up, abs_usize, vp, abs_vsize);
    }
  }

  w->_mp_size = wsize;
  if (delalloc != 0)
    mmpFree (del, delalloc * BYTES_PER_MP_LIMB);
}

/*3
* a[i] = a[i] + b[i]*x
*/
static void nlAddMult (mp_ptr a, mp_srcptr b, const int l, const unsigned int x)
{
  int i=1;
  _Int_64 v,w;
  
  mul_64(&v,*b,x);
  *a += v.lo;
  if (*a<v.lo) v.hi++;
  for (;;) {
    if (i==l)
    {
      a[l] = v.hi;
      return;
    }
    mul_64(&w,b[i],x);
    w.lo += v.hi;
    if (w.lo<v.hi) w.hi++;
    a[i] += w.lo;
    if (a[i]<w.lo) w.hi++;
    i++;
    if (i==l)
    {
      a[l] = w.hi;
      return;
    }
    mul_64(&v,b[i],x);
    v.lo += w.hi;
    if (v.lo<w.hi) v.hi++;
    a[i] += v.lo;
    if (a[i]<v.lo) v.hi++;
    i++;
  }
}

/*3
* a[i] = b[i]*x
*/
static void nlIniMult (mp_ptr a, mp_srcptr b, const int l, const unsigned int x)
{
  int i=1;
  _Int_64 v,w;
  
  mul_64(&v,*b,x);
  *a = v.lo;
  for (;;) {
    if (i==l)
    {
      a[l] = v.hi;
      return;
    }
    mul_64(&w,b[i],x);
    w.lo += v.hi;
    if (w.lo<v.hi) w.hi++;
    a[i] = w.lo;
    i++;
    if (i==l)
    {
      a[l] = w.hi;
      return;
    }
    mul_64(&v,b[i],x);
    v.lo += w.hi;
    if (v.lo<w.hi) v.hi++;
    a[i] = v.lo;
    i++;
  }
}

/*3
* aq := aq div b, b small
*/
static mp_limb_t nlDivModSmall (mp_ptr qp, mp_srcptr np, int nsize,
const unsigned int b)
{
  unsigned int hi,lo,q,c,inc;
  
  lo = np[--nsize];
  q = lo/b;
  if (q || (qp==np))
    qp[nsize] = q;
  lo -= b*q;
  if (nsize == 0)
    return lo;
  if (b<0x10000)
  {
    do
    {
      hi = lo;
      lo = np[--nsize];
      lo = divmod_ll_s(&q, hi, lo, b);
      qp[nsize] = q;
    } while (nsize);
    return lo;
  }
  c = b;
  if (inc = lead_zero(c))
  {
    c <<= inc;
    lo <<= inc;
    lo += np[nsize-1]>>(32-inc);
  }
  do
  {
    hi = lo;
    lo = np[--nsize];
    if (inc)
    {
      lo <<= inc;
      if (nsize)
        lo += np[nsize-1]>>(32-inc);
    }
    lo = divmod_ll_u(&q, hi, lo, c);
    qp[nsize] = q;
  } while (nsize);
  return lo>>inc;
}

/*3
* r := rp mod b, b small
* nsize = length(rp)
*/
static mp_limb_t nlRemSmall (mp_srcptr rp, int nsize, const unsigned int b)
{
  unsigned int hi,lo,c,inc;
  
  lo = rp[--nsize]%b;
  if (nsize == 0)
    return lo;
  if (b<0x10000)
  {
    do
    {
      hi = lo;
      lo = rp[--nsize];
      lo = rem_ll_s(hi, lo, b);
    } while (nsize);
    return lo;
  }
  c = b;
  if (inc = lead_zero(c))
  {
    c <<= inc;
    lo <<= inc;
    lo += rp[nsize-1]>>(32-inc);
  }
  do
  {
    hi = lo;
    lo = rp[--nsize];
    if (inc)
    {
      lo <<= inc;
      if (nsize)
        lo += rp[nsize-1]>>(32-inc);
    }
    lo = rem_ll_u(hi, lo, c);
  } while (nsize);
  return lo>>inc;
}


/*3
* aq := aq div bn, bn long
*/
static int  nlDivModLong(mp_ptr quot, const int ql,
mp_ptr ax, int al, mp_ptr bx, const int bl)
{
  const unsigned int inc = leading_zero(bx[bl-1]);
  mp_ptr an, bn;
  unsigned int hi, div, qb, q;

  
  al -= bl;
  an = ax + (al-1);
  bn = bx - 1;
  qb = 0;
  if (inc)
  {
    hi = an[bl]>>(32-inc);
    if (!hi)
    {
      if (ql > al)
      {
        qb++;
        hi = nlSubSimple(an, bn, bl);
        if (!al)
        {
          quot[al] = qb;
          return nlShrink0(an, bl);
        }
      }
    }
    nlLShift(ax, al+bl, inc);
    nlLShift(bx, bl, inc);
    div = bn[bl]+1;
    while (hi)
    {
      q = div_ll_u(hi, an[bl], div);
      qb += q;
      if (q>1)
        hi -= nlSubMult(an, bn, q, bl);
      else
        hi -= nlSubSimple(an, bn, bl);
    }
  }
  else
    div = bn[bl]+1;
  if (nlCompHigh(an, bn, bl))
  {
    qb++;
    hi = nlSubSimple(an, bn, bl);
  }
  if (qb)
    quot[al] = qb;
  for(;;)
  {
    if (!al)
    {
      if (inc)
      {
        nlRShift(ax, bl, inc);
        nlRShift(bx, bl, inc);
      }
      return nlShrink0(an, bl);
    }
    al--;
    an--;
    qb = 0;
    hi = an[bl+1];
    while (hi)
    {
      q = div_ll_u(hi, an[bl], div);
      qb += q;
      if (q>1)
        hi -= nlSubMult(an, bn, q, bl);
      else
        hi -= nlSubSimple(an, bn, bl);
    }
    if(nlCompHigh(an, bn, bl))
    {
      qb++;
      hi = nlSubSimple(an, bn, bl);
    }
    quot[al] = qb;
  }
}

/*3
* rl := an mod bn, bn long, but <= an
* al = length(an)
*/
static int  nlRemLong(mp_ptr ax, mp_ptr bx, int al, const int bl)
{
  const unsigned int inc = leading_zero(bx[bl-1]);
  mp_ptr an, bn;
  unsigned int hi, div, q;

  
  al -= bl;
  an = ax + (al-1);
  bn = bx - 1;
  if (inc)
  {
    hi = an[bl]>>(32-inc);
    if (!hi)
    {
      if (nlCompHigh(an, bn, bl))
      {
        hi = nlSubSimple(an, bn, bl);
        if (!al)
          return nlShrink0(an, bl);
      }
    }
    nlLShift(ax, al+bl, inc);
    nlLShift(bx, bl, inc);
    div = bn[bl]+1;
    while (hi)
    {
      q = div_ll_u(hi, an[bl], div);
      if (q>1)
        hi -= nlSubMult(an, bn, q, bl);
      else
        hi -= nlSubSimple(an, bn, bl);
    }
  }
  else
    div = bn[bl]+1;
  if (nlCompHigh(an, bn, bl))
    hi = nlSubSimple(an, bn, bl);
  for(;;)
  {
    if (!al)
    {
      if (inc)
      {
        nlRShift(ax, bl, inc);
        nlRShift(bx, bl, inc);
      }
      return nlShrink0(an, bl);
    }
    al--;
    an--;
    q = 0;
    hi = an[bl+1];
    while (hi)
    {
      q = div_ll_u(hi, an[bl], div);
      if (q>1)
        hi -= nlSubMult(an, bn, q, bl);
      else
        hi -= nlSubSimple(an, bn, bl);
    }
    if(nlCompHigh(an, bn, bl))
      hi = nlSubSimple(an, bn, bl);
  }
}

/*3
* z = GCD(a, b), a > b
*/
static void nlGcdLong (mp_ptr x, int *zl, mp_ptr y, int yl,
mp_ptr *z)
{
  mp_ptr h;
  int  hl, xl = *zl;
  mp_limb_t xc, yc, hc;

  if (xl == 1) {
    xc = *x;
    yc = *y;
    goto small;
  }
  for (; ; ) {
    if (yl == 1)
    {
      yc = *y;
      hc = (mp_limb_t)nlRemSmall(x, xl, yc);
      if (hc == 0) {
        *z = y;
        *zl = 1;
        return;
      }
      break;
    }
    hl = nlRemLong(x, y, xl, yl);
    if (hl == 0) {
      *z = y;
      *zl = yl;
      return;
    }
    h = x;
    x = y;
    y = h;
    xl = yl;
    yl = hl;
  }

  do {
    xc = yc;
    yc = hc;
small:
    hc = xc % yc;
  } while (hc != 0);
  *y = yc;
  *z = y;
  *zl = 1;
  return;
}

static void mpn_gcd (MP_INT *w, const MP_INT *u, const MP_INT *v)
{
  MP_INT t_u, t_v;
  MP_INT *p_u = &t_u, *p_v = &t_v;
  int usize, vsize, wsize;
  mp_ptr up, vp, wp;

  up = u->_mp_d;
  vp = v->_mp_d;
  wp = w->_mp_d;
  usize = ABS (u->_mp_size);
  vsize = ABS (v->_mp_size);

  /* GCD(u,0) == u, GCD(u,1) == 1  */
  if (vsize < 2) {
    if (vsize == 0) {
      mpz_set (w, u);
      w->_mp_size = usize;
      return;
    }
    if (vp[0] == 1) {
      wp[0] = 1;
      w->_mp_size = 1;
      return;
    }
  }

  /* make copy's */
  mpz_init_set(p_u, u);
  mpz_init_set(p_v, v);
  up = p_u->_mp_d;
  vp = p_v->_mp_d;

  /* GCD(u,v) */
  wsize = usize;
  nlGcdLong(up, &wsize, vp, vsize, &wp);
  mpz_clear(w);
  w->_mp_size = wsize;
  w->_mp_d = wp;
  if (wp != p_u->_mp_d) {
    mpz_clear(p_u);
    w->_mp_alloc = p_v->_mp_alloc;
  } else {
    mpz_clear(p_v);
    w->_mp_alloc = p_u->_mp_alloc;
  }
}

/*3
* a := a - b, b1...bj
*/
static mp_limb_t nlSubSimple (mp_ptr a, mp_srcptr b, const int l)
{
  int j = 1;
  mp_limb_t m;

  /* NON-CARRY LOOP */
  do {
    m = a[j];
    a[j] = m - b[j];
    if (a[j] > m)
      goto cy_loop;
ncy_loop:
    j++;
  } while (j <= l);
  return 0;

  /* CARRY LOOP */
  do {
    m = a[j];
    a[j] = m - (b[j] + 1);
    if (a[j] < m)
      goto ncy_loop;
cy_loop:
    j++;
  } while (j <= l);
  return 1;
}

/*3
* a := a - b*x, b1...bj
*/
static mp_limb_t nlSubMult (mp_ptr a, mp_srcptr b,
const unsigned int x, const int l)
{
  int j = 1;
  mp_limb_t m,h;
  _Int_64 y;

  mul_64(&y,b[1],x);
  h = y.hi;
  m = a[1];
  a[1] = m - y.lo;
  if (a[1] > m)
    h++;
  j = 2;

  do {
    mul_64(&y,b[j],x);
    y.lo += h;
    if (y.lo<h)
      y.hi++;
    h = y.hi;
    m = a[j];
    a[j] = m - y.lo;
    if (a[j] > m)
      h++;
    j++;
  } while (j <= l);
  return h;
}

/*3
* compare xi and yi from i=l to 1
*/
static int nlCompHigh (mp_srcptr x, mp_srcptr y, const int l)
{
  int  i;
  for (i = l; i != 0; i--) {
    if (x[i] != y[i])
      goto diff;
  }
  return 1;
diff:
  return x[i] < y[i] ? 0 : 1;
}

static int  nlShrink0(mp_srcptr c, int j)
{
  while ((j != 0) && (c[j] == 0)) 
    j--;
  return j;
}

/* MIN > SUB > 0 */
static int  mpn_sub (mp_ptr dif_ptr,
mp_srcptr min_ptr, int min_size,
mp_srcptr sub_ptr, int sub_size)
{
  int j = 0;
  mp_limb_t m, dif;

  /* There are two do-loops, marked NON-CARRY LOOP and CARRY LOOP that
     jump between each other.  The first loop is for when the previous
     subtraction didn't produce a carry-out; the second is for the
     complementary case.  */

  /* NON-CARRY LOOP */
  do {
    m = min_ptr[j];
    dif = dif_ptr[j] = m - sub_ptr[j];
    if (dif > m)
      goto cy_loop;
ncy_loop:
    j++;
  } while (j < sub_size);
  if (j == min_size)
    goto normalize;

  /* We have exhausted SUB, with no carry out.  Copy remaining part of
     MIN to DIF.  */
  do {
    dif_ptr[j] = min_ptr[j];
  } while (++j < min_size);
  return j;

  /* CARRY LOOP */
  do {
    m = min_ptr[j];
    dif = dif_ptr[j] = m - (sub_ptr[j] + 1);
    if (dif < m)
      goto ncy_loop;
cy_loop:
    j++;
  } while (j < sub_size);

  /* We have exhausted SUB, but need to propagate carry.  */
  do {
    m = min_ptr[j];
    dif_ptr[j] = dif = m - 1;
    j++;
  } while (m == 0);
  if (j == min_size)
    goto normalize;

  do {
    dif_ptr[j] = min_ptr[j];
  } while (++j < min_size);
  return j;

normalize:
  if (dif != 0)
    return j;
  for (; ; ) {
    j--;
    if (dif_ptr[j-1] != 0)
      return j;
  }
}

/* length(ADD1) >= length(ADD2) >= 0, length(ADD1) != 0 */
static int  mpn_add (mp_ptr sum_ptr,
mp_srcptr add1_ptr, int add1_size,
mp_srcptr add2_ptr, int add2_size)
{
  int j = 0;
  mp_limb_t a1, sum;

  if (add2_size == 0)
    goto copy;

  /* There are two do-loops, marked NON-CARRY LOOP and CARRY LOOP that
     jump between each other.  The first loop is for when the previous
     addition didn't produce a carry-out; the second is for the
     complementary case.  */

  /* NON-CARRY LOOP */
  do {
    a1 = add1_ptr[j];
    sum = sum_ptr[j] = a1 + add2_ptr[j];
    if (sum < a1)
      goto cy_loop;
ncy_loop:
    j++;
  } while (j < add2_size);
  if (j < add1_size)
    goto copy;
  return j;

  /* CARRY LOOP */
  do {
    a1 = add1_ptr[j];
    sum = sum_ptr[j] = a1 + add2_ptr[j] + 1;
    if (sum > a1)
      goto ncy_loop;
cy_loop:
    j++;
  } while (j < add2_size);
  if (j == add1_size) {
    sum_ptr[j] = 1;
    return j + 1;
  }

  /* We have exhausted ADD2, but need to propagate carry.  */
  for (; ; ) {
    sum = sum_ptr[j] = add1_ptr[j] + 1;
    j++;
    if (sum != 0) {
      if (j == add1_size)
        return j;
      break;
    }
    if (j == add1_size) {
      sum_ptr[j] = 1;
      return j + 1;
    }
  }

copy:
  do {
    sum_ptr[j] = add1_ptr[j];
  } while (++j < add1_size);
  return j;
}

static void nlLShift(mp_ptr ax,
const unsigned int al, const unsigned int inc)
{
  int i=al-1, coinc=32-inc;
  unsigned int hi=ax[i]<<inc, lo;
  
  while (i)
  {
    i--;
    lo = ax[i];
    ax[i+1] = hi+(lo>>coinc);
    hi = lo<<inc;
  }
  *ax = hi;
}

static void nlRShift(mp_ptr ax,
const unsigned int al, const unsigned int inc)
{
  int i=1, coinc=32-inc;
  unsigned int lo=*ax>>inc, hi;
  
  while (i<al)
  {
    hi = ax[i];
    ax[i-1] = lo+(hi<<coinc);
    lo = hi>>inc;
    i++;
  }
  ax[i-1] = lo;
}
